# Author: Qi Li
# command
# python mBPIs.py -in <INPUT STATIC NETWORK> -dim <DESIRED FEATURE DIMENSION>
# example 
# python mBPIs.py -in ./Data/HPRD_Subnetworks/NetWalk-Static.txt -dim 30

import numpy as np
import pandas as pd
import sys 
from os import listdir
import networkx as nx
from collections import defaultdict
import statistics as st
import argparse
import os

##################################################
if not os.path.exists('./output/'):
    os.makedirs('./output/')

def parse_args():
    '''
    Parses arguments 
    '''

    parser = argparse.ArgumentParser(description="Run feature mBPIs")

    parser.add_argument("-in", "--input-network", dest='fn', type=str,
                        help="Input network file")

    return parser.parse_args()


def Merge(dict1, dict2): 
    res = {**dict1, **dict2} 
    return (res)  


### concatenation of the feature ####
def Concatenation(df1, df2):
    df = pd.merge(df1, df2, how = 'left', on = ['node'])
    return (df)

### calculate degree, aging neighbor, aging neighbor ratio ###
def Centrality_static(fn):
    df = pd.read_csv(fn, header = None, sep = '\t',  dtype={0: 'str', 1: 'str'})

    node = list(set(df.iloc[:, 0].tolist() + df.iloc[:, 1].tolist()))
    G = nx.read_edgelist(fn)
    node = list(set(list(G.nodes)))

    ### neighbor start ###
    neighbor = {}
    for i in node:
        neighbor[i] = []

    for i in range(df.shape[0]):
        neighbor[df.iloc[i, 0]].append(df.iloc[i, 1])
        neighbor[df.iloc[i, 1]].append(df.iloc[i, 0])

    L1 = []
    for i in node:
        L = list(set(neighbor[i]))
        L1.append(len(L))

    df0 = pd.DataFrame(data = {'node': node, 'NO_neighbor': L1})

    #df0.to_csv('../UniNet/Raw/' + net + '_' + pos.split(".")[0] + '_Neighbor.txt', sep = '\t', header = None, index = False)
    ### neighbor end ###


    ### K-core start ###
    k = nx.core_number(G)
    df1 = pd.DataFrame(data = {'node': list(k.keys()), 'K-core': list(k.values())})
    #df1.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_K-core.txt', sep = '\t', header = None, index = False)
    ### K-core end ###

    ### Average shortest path start ###
    sh = dict(nx.shortest_path_length(G))
    node1 = []
    path_L = []
    for i in sh.keys():
        node1.append(i)
        path_L.append(st.mean(list(sh[i].values())))
    df2 = pd.DataFrame(data = {'node': node1, 'shortestPath': path_L})
    #df2.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_Avg_shortest_path.txt', sep = '\t', header = None, index = False)
    ### Average shortest path end ###

    ### Betweenness Centrality star ###
    B = nx.betweenness_centrality(G, normalized=True, weight=None, endpoints=False, seed=None)
    df3 = pd.DataFrame(data = {'node': list(B.keys()), 'betweenness': list(B.values())})
    #df3.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_Betweenness.txt', sep = '\t', header = None, index = False)
    ### Betweenness Centrality end ###

    ### Closeness centrality start ###
    c = nx.closeness_centrality(G)
    df4 = pd.DataFrame(data = {'node': list(c.keys()), 'closeness': list(c.values())})
    #df4.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_closeness.txt', sep = '\t', header = None, index = False)
    ### Closeness centrality end ###

    ### Clustering coefficient start ###
    cc = nx.clustering(G)
    df5 = pd.DataFrame(data = {'node': list(cc.keys()), 'clusteringcoefficient': list(cc.values())})
    #df5.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_clustering_coefficient.txt', sep = '\t', header = None, index = False)
    ### Clustering coefficient end ###

    ### Eccentricity start ###
    ecc = {}
    for c in nx.connected_components(G):
        c1 = G.subgraph(c).copy()
        ecc = Merge(ecc, nx.eccentricity(c1))
    df6 = pd.DataFrame(data = {'node': list(ecc.keys()), 'ecc': list(ecc.values())})
    #df6.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_Eccentricity.txt', sep = '\t', header = None, index = False)
    ### Eccentricity end ###
    
    ### Neighborhood Connectivity start ###
    # output: A dictionary keyed by degree k with the value of average connectivity.
    # in this case, we need to map back the degree to nodes ##
    nodedegree = G.degree
    V, D = [], []
    for v, d in nodedegree:
        V.append(v)
        D.append(d)

    df7_0 = pd.DataFrame(data = {'node':V, 'degree': D})
    nc = nx.average_degree_connectivity(G)
    df7 = pd.DataFrame(data = {'degree': list(nc.keys()), 'NC': list(nc.values())})

    df7_0 = pd.merge(df7_0, df7, how = "left", on = ['degree'])
    df7 = df7_0[['node', 'NC']]
    #df7.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_neighbor_connectivity.txt', sep = '\t', header = None, index = False)
    ### Neighborhood Connectivity end ###

    ### Radiality start ###
    # radiality(v) is calculated by computing the shortest path between the node v and all other nodes in the graph. 
    # The value of each path is then subtracted by the value of the diameter + 1 and the resulting values are summated. 
    # Finally, the obtained value is divided for the number of nodes -1 (n-1). 

    node1 = []
    radia = []

    for C in nx.connected_components(G):
        tmp = dict(nx.shortest_path_length(G))
        C1 = G.subgraph(C).copy()
        d = nx.diameter(C1) + 1
        node_tmp = list(C1.nodes)
        radia_tmp = []
        n = len(list((C1.nodes)))
        for i in node_tmp:
            tmp_L = list(tmp[i].values())
            tmp_L = [d - int(x) for x in tmp_L]
            radia_tmp.append(sum(tmp_L)*1.0/(n-1))
        node1 = node1 + node_tmp
        radia = radia + radia_tmp

    df8 = pd.DataFrame(data = {'node': node1, 'radiality':radia})
    #df8.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_Radiality.txt', sep = '\t', header = None, index = False)
    ### Radiality end ###
    
    ### Stress start ###
    def num_spaths(G):
        n_spaths = dict.fromkeys(G, 0.0)
        spaths = dict(nx.all_pairs_shortest_path(G))

        for source in G:
            for path in spaths[source].values():
                for node in path[1:]: # ignore first element (source == node)
                    n_spaths[node] += 1 # this path passes through `node`

        return (n_spaths)

    #spaths = nx.all_pairs_shortest_path(G)
    n_spaths = num_spaths(G)

    df9 = pd.DataFrame(data = {'node': list(n_spaths.keys()), 'stress': list(n_spaths.values())})
    #df9.to_csv('../UniNet/' + net + '_' + pos.split(".")[0] + '_Stress.txt', sep = '\t', header = None, index = False)
    ### Stress end ###

    #files = [df2, df3, df6, df4, df5, df1,df0,df8,df9,df7]    

    # Centrality Concatenation #
    df_uninet = pd.read_csv("./topo_coefficient.txt", header = None, sep = "\t",  dtype={0: 'str'})
    df_uninet.columns = ['node', 'topo_coefficient']

    df_uninet = Concatenation(df_uninet, df2)
    df_uninet = Concatenation(df_uninet, df3)
    df_uninet = Concatenation(df_uninet, df6)
    df_uninet = Concatenation(df_uninet, df4)
    df_uninet = Concatenation(df_uninet, df5)
    df_uninet = Concatenation(df_uninet, df1)
    df_uninet = Concatenation(df_uninet, df0)
    df_uninet = Concatenation(df_uninet, df8)
    df_uninet = Concatenation(df_uninet, df9)
    df_uninet = Concatenation(df_uninet, df7)

    netn = fn.split("/")[-1].split(".txt")[0]
    df_uninet.to_csv("./output/UniNet_" + netn + ".txt", header = None, sep = '\t', index = False)

##################################################################################################
if __name__ == "__main__":

    args = parse_args()
    Centrality_static(args.fn)



